/*
 * Copyright (C) 2020 Richard Hughes <richard@hughsie.com>
 *
 * SPDX-License-Identifier: LGPL-2.1+
 */

#pragma once

#include "jcat-context.h"

GPtrArray *
jcat_context_get_public_keys(JcatContext *self) G_GNUC_NON_NULL(1);
